/*
 *  Common include file for all Plugin DB sources
 */

#ifndef PLUGINDB_H
#define PLUGINDB_H


// precompiled portion
#include "plugindb_precomp.h"

// other common includes
#include "sysync.h"

// Items used by ODBC DB interface
#include "multifielditem.h"
#include "mimediritemtype.h"

#endif

/* eof */
