/*
 *  File:         platform_headers.h
 *
 *  Author:       Lukas Zeller (luz@plan44.ch)
 *
 *  Common include files for all platform-related standard headers
 *  (suitable for precompiled headers and/or prefix file)
 *
 *  Copyright (c) 2003-2011 by Synthesis AG + plan44.ch
 *
 *  NOTE: this file is part of the mod_sysync source files.
 *
 */


#ifndef __PLATFORM_HEADERS_H
#define __PLATFORM_HEADERS_H

// ANSI C

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>

// C++
#ifdef __cplusplus


#endif // __cplusplus


#endif

// eof
