/*
 *  File:         profiling.cpp
 *
 *  Author:			  Lukas Zeller (luz@plan44.ch)
 *
 *  Platform-dependent profiling implementation
 *
 *  Copyright (c) 2002-2011 by Synthesis AG + plan44.ch
 *
 *  2002-10-01 : luz : created
 *
 */

#include "profiling.h"
#include "sysync_globs.h"

#ifdef TIME_PROFILING

#error "No time profiling for linux yet"

#endif // TIME_PROFILING


#ifdef MEMORY_PROFILING

#error "No memory  profiling for linux yet"

#error "Seems to be dangerous"

#endif // MEMORY_PROFILING

/* eof */
