/* client engine (w/o internal transport) generic header file
 *
 */

#ifndef ENGINE_CLIENT_H
#define ENGINE_CLIENT_H

/* Headers that might be available in precompiled form
 * (standard libraries, SyncML toolkit...)
 */
#include "engine_client_precomp.h"

/* headers not suitable for / entirely included in precompilation */

// sysync core
#include "sysync.h"
// platform utilities
//#include "platform_utils.h"

// classes
#include "engineclientbase.h"

/* globals */


#endif // ENGINE_CLIENT_H

/* eof */
