/*
 *  precompiled/precompilable headers
 */

#ifndef PLUGINDB_PRECOMP_H
#define PLUGINDB_PRECOMP_H


/* Specific headers for Plugin DB interface
 */

#ifdef _WIN32
  #include <windows.h>
#endif

#endif

/* eof */
