/* C Headers that might be available in precompiled form
 * for SML toolkit compilation.
 * (standard libraries, SyncML toolkit itself...)
 */

#ifndef SMLTK_PRECOMP_XPT_H
#define SMLTK_PRECOMP_XPT_H


/* SyncML Toolkit includes */
// standard stuff
#include "smltk_precomp.h"
// plus XPT stuff
#include "xpt.h"


#endif /* defined SMLTK_PRECOMP_XPT_H */
