EXPORTS
smlAddCmd
smlAlertCmd
smlAllocAdd
smlAllocAlert
smlAllocAtomic
smlAllocChal
smlAllocCopy
smlAllocCred
smlAllocDelete
smlAllocDevInfCTCap
smlAllocDevInfCTData
smlAllocDevInfCTDataList
smlAllocDevInfCTDataProp
smlAllocDevInfCTDataPropList
smlAllocDevInfCtcapList
smlAllocDevInfDSMem
smlAllocDevInfDatastore
smlAllocDevInfDatastoreList
smlAllocDevInfDevInf
smlAllocDevInfExt
smlAllocDevInfExtList
smlAllocDevInfSyncCap
smlAllocDevInfXmit
smlAllocDevInfXmitList
smlAllocExec
smlAllocGeneric
smlAllocGet
smlAllocItem
smlAllocItemList
smlAllocMap
smlAllocMapItem
smlAllocMapItemList
smlAllocMetInfAnchor
smlAllocMetInfMem
smlAllocMetInfMetInf
smlAllocPcdata
smlAllocPcdataList
smlAllocPut
smlAllocReplace
smlAllocResults
smlAllocSearch
smlAllocSequence
smlAllocSource
smlAllocSourceList
smlAllocSourceRefList
smlAllocStatus
smlAllocSync
smlAllocSyncHdr
smlAllocTarget
smlAllocTargetRefList
smlAllocUnknownProtoElement
smlCopyCmd
smlDeleteCmd
smlEndAtomic
smlEndMessage
smlEndSequence
smlEndSync
smlExecCmd
smlFreeAlert
smlFreeAtomic
smlFreeChalPtr
smlFreeCredPtr
smlFreeDevInfCTCap
smlFreeDevInfCTData
smlFreeDevInfCTDataList
smlFreeDevInfCTDataProp
smlFreeDevInfCTDataPropList
smlFreeDevInfCtcapList
smlFreeDevInfDSMem
smlFreeDevInfDatastore
smlFreeDevInfDatastoreList
smlFreeDevInfDevInf
smlFreeDevInfExt
smlFreeDevInfExtList
smlFreeDevInfSynccap
smlFreeDevInfXmit
smlFreeDevInfXmitList
smlFreeExec
smlFreeGeneric
smlFreeGetPut
smlFreeItemList
smlFreeItemPtr
smlFreeMap
smlFreeMapItemList
smlFreeMapItemPtr
smlFreeMetinfAnchor
smlFreeMetinfMem
smlFreeMetinfMetinf
smlFreePcdata
smlFreePcdataList
smlFreeProtoElement
smlFreeResults
smlFreeSearch
smlFreeSourceList
smlFreeSourceRefList
smlFreeSourceTargetPtr
smlFreeStatus
smlFreeSync
smlFreeSyncHdr
smlFreeTargetRefList
smlGetCmd
smlGetFreeBuffer
smlInit
smlInitInstance
smlLibFree
smlLibMalloc
smlLibMemcmp
smlLibMemcpy
smlLibMemmove
smlLibMemset
smlLibPrint
smlLibRealloc
smlLibStrcat
smlLibStrchr
smlLibStrcmp
smlLibStrcpy
smlLibStrdup
smlLibStrlen
smlLibStrncat
smlLibStrncmp
smlLibStrncpy
smlLibStrstr
smlLibVprintf
smlLockReadBuffer
smlLockWriteBuffer
smlMapCmd
smlPcdata2String
smlPcdataDup
smlProcessData
smlPutCmd
smlReplaceCmd
smlResultsCmd
smlSearchCmd
smlSetCallbacks
smlSetEncoding
smlSetSyncMLOptions
smlSetUserData
smlStartAtomic
smlStartMessage
smlStartMessageExt
smlStartSequence
smlStartSync
smlStatusCmd
smlString2Pcdata
smlTerminate
smlTerminateInstance
smlUnlockReadBuffer
smlUnlockWriteBuffer
smlStartEvaluation
smlEndEvaluation
