#ifndef __SMLTK_PREFIX_FILE_H
#define __SMLTK_PREFIX_FILE_H

// C version use
// - target options can incfluence everything
#include "target_options.h"
// - SML toolkit WITHOUT xpt part
#include "smltk_precomp.h"

#endif